"""Noojing
"""


class Portfolio:
    def __init__(self, weights=None, tickers=None) -> None:
        self._weights = weights
        self._tickers = tickers
        self._timeseries = None

    @property
    def tickers(self):
        return self._tickers

    @tickers.setter
    def tickers(self, newtickers):
        if (
            self._tickers is None
            and self._weights is not None
            and len(self._weights) != len(newtickers)
        ):
            raise ValueError(
                "[tickers] definition should have the same shape as [weights]"
                f" - weights: {self._weights}; tickers: {newtickers}."
            )

        if self._tickers is not None and len(self._tickers) != len(newtickers):
            raise ValueError(
                "[tickers] redefinition should remain the same shape"
                f" - Old: {self._tickers}; new: {newtickers}."
            )

        self._tickers = newtickers

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, newweights):
        if (
            self._tickers is None
            and self._weights is not None
            and len(self._weights) != len(newtickers)
        ):
            raise ValueError(
                "[weights] definition should have the same shape as [tickers]"
                f" - tickers: {self._tickers}; weights: {newweights}."
            )

        if self._weights is not None and len(self._weights) != len(newweights):
            raise ValueError(
                "[weights] redefinition should remain the same shape"
                f" - Old: {self._weights}; new: {newweights}."
            )

        self._weights = newweights

    @property
    def timeseries(self):
        return self._timeseries

    @timeseries.setter
    def timeseries(self, newtimeseries):
        if self.tickers is None:
            self.tickers = list(newtimeseries.columns)
            self._timeseries = newtimeseries
            return

        if len(newtimeseries.columns) != len(self.tickers):
            raise ValueError(
                "[timeseries] definition should have the consistent shape as [tickers]"
                f" - tickers: {self._tickers}; timeseries: {newtimeseries.shape}."
            )

        if (self.tickers != newtimeseries.columns).all():
            self.tickers = list(newtimeseries.columns)

        self._timeseries = newtimeseries



from .baseportfolio import BasePortfolio

import numpy as np


class BlackLittermanPortfolio(BasePortfolio):
    """Class for modelling Black-Litterman portfolios

    Args:
        BasePortfolio (BasePortfolio): inheriting for some commonly used functions, e.g.: get_weights
    """

    def __init__(
        self,
        equilcovariances,
        equilweights,
        P,
        Q,
        riskaversion=None,
        tau=0.05,
        omega=None,  # specified values here is take as priority
    ):
        """Constructor for passing equilibrium parameters and views:

        Args:
            equilcovariances (np.array): covariance matrix from equilibrium portfolio
            equilweights (np.array): column vector for weights from equilibrium portfolio
            P (np.array): K x N matrix, where K is number of views and N is number of assets
            Q (np.array): K column vector for absolute or relative returns
            riskaversion (float, optional): if not defined, calculate a default in the constructor. Defaults to None.
            tau (float, optional): ratio of covariances between views and estimated returns. Defaults to 0.05.
            omega (np.array, optional): K x K matrix, which avoids hassle of estimating omega in model. Defaults to None.
        """

        # inheriting from base
        super().__init__()

        self._equil_covariances = equilcovariances
        self._equil_weights = equilweights

        # Initial portfolio: Sigma, weigh, risk aversion
        if riskaversion is None:
            # reference Idzorek(2002) A STEP-BY-STEP GUIDE TO THE BLACK-LITTERMAN MODEL
            self._risk_aversion = 0.5 / np.sqrt(
                (equilweights.T @ equilcovariances @ equilweights)[0, 0]
            )
        else:
            self._risk_aversion = riskaversion

        # views; P, v
        self._P = P
        self._Q = Q
        self._tau = tau

        self._pi = None
        self._omega = omega  # calculated

        return

    def get_covariances(self, **kwargs):
        """override function from base"""
        if self._covariances is None:
            self._get_bl_returns_covariancematrix(**kwargs)
        return self._covariances

    def get_weights(self, **kwargs):
        """Override function from base"""
        return self._get_bl_weights(**kwargs)

    def get_returns(self, **kwargs):
        """Override function from base"""
        if self._returns is None:
            self._get_bl_returns_covariancematrix(**kwargs)
        return self._returns

    def get_pi(self):
        """Estimate returns from the equilibrium portfolio"""
        if self._pi is None:
            self._pi = self._risk_aversion * (
                self._equil_covariances @ self._equil_weights
            )
        return self._pi

    def get_view_omega(self):
        """get covariances on views:
        for now, we simply take \tau P * \Sigma * P
        """
        if self._omega is None:
            self._omega = np.diagflat(
                np.diag(self._tau * self._P @ self._equil_covariances @ self._P.T)
            )
        return self._omega

    def get_risk_aversion(self):
        """Get risk aversion"""
        return self._risk_aversion

    def _get_bl_returns_covariancematrix(self, version):
        """Get updated covariance matrix and returns upon BL model:
        use different formula in Meucci(2010) - The Black-Litterman Approach: Original Model and Extensions

        Args:
            version (str): valid string - v0, v1 and v2, which can be seen from related private functions below

        Raises:
            ValueError: throws if no valid string for version

        Returns:
            tuple: BL returns and BL covariance matrix
        """
        if version == "v0":
            return self._get_bl_returns_covariancematrix_v0()
        elif version == "v1":
            return self._get_bl_returns_covariancematrix_v1()
        else:
            raise ValueError(
                f'Invalid value for {version} - only "v0" or "v1" are supported.'
            )

    def _get_bl_weights(self, version, riskaversion=None):
        """get BL weights:
        - use analytical solution for unconstrained optimization if calculating for the 1st time;
        - otherwise, use the cached one

        Args:
            version (str): valid string - v0, v1 and v2, which can be seen from related private functions below
            riskaversion (float, optional): update risk aversion if given here. Defaults to None.

        Returns:
            np.array: BL weights
        """
        # can specify different risk aversion?
        risk_aversion = self._risk_aversion if riskaversion is None else riskaversion
        if self._returns is None:
            self._get_bl_returns_covariancematrix(version)
        bl_weights = np.linalg.inv(self._covariances) @ self._returns / risk_aversion
        return bl_weights

    def _get_bl_returns_covariancematrix_v0(self):
        """Get BL returns and covariance by Meucci(2010):
        - eq(16) for returns;
        - eq(17) for covariance matrix

        Returns:
            tuple: BL returns and BL covariance matrix
        """
        if self._returns is None or self._covariances is None:
            a = np.linalg.inv(self._tau * self._equil_covariances)
            omega_inv = np.linalg.inv(self.get_view_omega())
            # eq(17)
            self._covariances = np.linalg.inv(a + self._P.T @ omega_inv @ self._P)
            # eq(16)
            self._returns = self._covariances @ (
                a @ self.get_pi() + self._P.T @ omega_inv @ self._Q
            )
        return self._returns, self._covariances

    def _get_bl_returns_covariancematrix_v1(self):
        """Get BL returns and covariance by Meucci(2010):
        - eq(20) for returns;
        - eq(21) for covariance matrix

        Returns:
            tuple: BL returns and BL covariance matrix
        """
        if self._returns is None or self._covariances is None:
            temp_mat = (
                self._equil_covariances
                @ self._P.T
                @ np.linalg.inv(
                    self._tau * self._P @ self._equil_covariances @ self._P.T
                    + self.get_view_omega()
                )
            )
            # eq(20)
            self._returns = self.get_pi() + self._tau * temp_mat @ (
                self._Q - self._P @ self.get_pi()
            )
            # eq(21)
            self._covariances = (1 + self._tau) * self._equil_covariances - np.square(
                self._tau
            ) * temp_mat @ self._P @ self._equil_covariances
        return self._returns, self._covariances

    def _get_bl_returns_covariancematrix_v2(self):
        """Get BL returns and covariance by Meucci(2010):
        - eq(32) for returns;
        - eq(33) for covariance matrix

        Returns:
            tuple: BL returns and BL covariance matrix
        """
        p_sig = self._P @ self._equil_covariances
        p_sig_p_omega = p_sig @ self._P.T + self.get_view_omega()
        p_sig_p_omega_inv = np.linalg.inv(p_sig_p_omega)

        # eq(32)
        self._returns = self.get_pi() + p_sig.T @ p_sig_p_omega_inv @ (
            self._Q - self._P @ self.get_pi()
        )
        # eq(33)
        self._covariances = self._covariances - p_sig.T @ p_sig_p_omega_inv @ p_sig

        return self._returns, self._covariances



from ..utils.show import verbose_print

from pathlib import Path

import yfinance as yf
import numpy as np
import pandas as pd

import concurrent.futures

# The by-default path directory
_md_cache_dir = Path.home() / "noofolio/md_cache"


def get_md_observable(
    symbols,
    observable="Adj Close",
    startdate=None,
    enddate=None,
    cachedir=_md_cache_dir,
    frequency="1D",
):
    """Get historical time series for a particular market observable:
    - delegate data sourcing to _get_daily_md_observable;
    - and then transform by frequency

    Args:
        symbols (array-like): a list of valid symbols
        observable (str, optional): Valid market observable. Defaults to 'Adj Close'.
        startdate (str|datetime, optional): start of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        enddate (str|datetime, optional): end of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        cachedir (Path, optional): directory where market data is cached. Defaults to _md_cache_dir.
        frequency (str, optional): available ones - 1D, 1W, 1M, 3M and 1Y. Defaults to '1D'.

    Raises:
        ValueError: throws if frequency is not recognized

    Returns:
        pd.DataFrame: time on rows and symbols on columns
    """
    md = _get_daily_md_observable(symbols, observable, startdate, enddate, cachedir)
    freq_upper = frequency.upper()
    if freq_upper == "1D":
        return md

    if freq_upper == "1W":
        # the last incomplete week of year will be put into the 1 week of new
        # year, so need special handling here
        md["GroupKey"] = [
            (
                i.year + 1 if i.month == 12 and i.isocalendar()[1] < 2 else i.year,
                i.isocalendar()[1],
            )
            for i in md.index
        ]
        return md.groupby(md["GroupKey"]).tail(1).drop(columns="GroupKey")

    if freq_upper == "1M":
        return md.groupby([md.index.year, md.index.month]).tail(1)

    if freq_upper == "3M":
        return md.groupby([md.index.year, md.index.quarter]).tail(1)

    if freq_upper == "1Y":
        return md.groupby(md.index.year).tail(1)

    raise ValueError(
        f"Unsupported value for parameter [frequency]: {frequency}. \
        Valid options - 1D, 1W, 1M, 3M and 1Y"
    )


def get_md_rfr(startdate=None, enddate=None, frequency="1Y"):
    """Get historical time series for risk-free rates:
    - currently treasury 3M rates manually sourcing from https://fred.stlouisfed.org/series/TB3MS and cached;
    - simple transformation by frequency

    Args:
        startdate (str|datetime, optional): start of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        enddate (str|datetime, optional): end of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        frequency (str, optional): available ones - 1D, 1W, 1M, 3M and 1Y. Defaults to '1Y'.

    Raises:
        ValueError: throws if frequency is not recognized

    Returns:
        pd.DataFrame: time and rows and a single column for the rate
    """
    rfr_path = Path(__file__).parent / "datastore/rfr_treasury3m.csv"
    rfr = pd.read_csv(rfr_path, index_col="DATE")
    rfr.index = pd.to_datetime(rfr.index)
    if startdate is not None:
        rfr = rfr[rfr.index >= startdate]
    if enddate is not None:
        rfr = rfr[rfr.index <= enddate]
    rfr = rfr / 100

    freq_upper = frequency.upper()
    if freq_upper == "1D":
        return rfr / 360

    if freq_upper == "1W":
        return rfr * 7 / 360

    if freq_upper == "1M":
        return rfr * 30 / 360

    if freq_upper == "3M":
        return rfr * 90 / 360

    if freq_upper == "1Y":
        return rfr

    raise ValueError(
        f"Unsupported value for parameter [frequency]: {frequency}. \
        Valid options - 1D, 1W, 1M, 3M and 1Y"
    )


def get_asset_info():
    """Get asset related attributes:
    - currently manually sourcing from https://www.nasdaq.com/market-activity/stocks/screener, as of 13th Jan,2021 and get it cached

    Returns:
        pd.DataFrame: each column is an associated attribute of an asset
    """
    symbols_path = Path(__file__).parent / "datastore/symbols.csv"
    return pd.read_csv(symbols_path)


def get_symbols_from_cache(cachedir=_md_cache_dir):
    """Extact out list of symbols from cached data:

    Args:
        cachedir (Path, optional): directory where market data is cached. Defaults to _md_cache_dir.

    Returns:
        array: list of asset symbols
    """
    paths = list(Path(cachedir).rglob("*.csv"))
    symbols = []
    for path in paths:
        if path.is_file():
            symbol = (
                path.name.replace(".csv", "").replace("md_", "").replace("%2f", "/")
            )
            symbols.append(symbol)
    return symbols


def precache_md(symbols, cachedir=_md_cache_dir, verbose=True):
    """Batch cache market data for later usage so to avoid:
    - avoid internet connection stability issue;
    - improve performance for avoiding repeated remote downloading

    Args:
        symbols (array-like): a list of valid symbols
        cachedir (Path, optional): directory where market data is cached. Defaults to _md_cache_dir.
        verbose (bool, optional): print out some more details on retrieval and progress. Defaults to True.

    Returns:
        int: number of symbols successfully cached
    """
    symbol_count = len(symbols)
    print(f"It's in total {symbol_count} symbols to download.")

    cached_count = 0
    for i in range(symbol_count):
        verbose_print(verbose, f"Resolving {i+1}/{symbol_count} {symbols[i]}...")
        try:
            s_encoded = symbols[i].replace("/", "%2f").lower()
            file_path = cachedir / f"md_{s_encoded}.csv"
            if file_path.exists():
                verbose_print(verbose, f"{symbols[i]} already exists, so skip.")
                continue
            data = yf.download(symbols[i], threads=False, progress=verbose)
            if data is not None and not data.empty:
                data.to_csv(file_path)
                verbose_print(verbose, f"{symbols[i]} saved.")
                cached_count += 1
        except Exception as ex:
            print(ex)
    return cached_count


def _get_daily_md_observable(
    symbols, observable, startdate=None, enddate=None, cachedir=_md_cache_dir
):
    """Get historical daily time series for a particular market observable:
    - sourced from Yahoo Finance only for now;
    - with cached data taken as priority

    Args:
        symbols (array-like): a list of valid symbols
        observable (str, optional): Valid market observable. Defaults to 'Adj Close'.
        startdate (str|datetime, optional): start of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        enddate (str|datetime, optional): end of historical period of format 'yyyy-mm-dd' if str. Defaults to None.
        cachedir (Path, optional): directory where market data is cached. Defaults to _md_cache_dir.

    Returns:
        pd.DataFrame: time on rows and symbols on columns
    """
    symbols = [symbols] if isinstance(symbols, str) else symbols

    if cachedir is not None and not cachedir.exists():
        cachedir.mkdir(parents=True)

    def _retrieve_md(symbol):  # , observable, startdate, enddate, cachedir
        file_path = None if cachedir is None else cachedir / f"md_{symbol}.csv".lower()
        md = None
        if file_path is not None and file_path.exists():
            md = pd.read_csv(file_path, index_col="Date")
        else:
            md = yf.download(symbol)
            if file_path is not None:
                md.to_csv(file_path)
        if md.empty:
            return None
        md.index = pd.to_datetime(md.index)
        md = md[[observable]]
        md.columns = [symbol.upper()]
        if startdate is not None:
            md = md[md.index >= startdate]
        if enddate is not None:
            md = md[md.index <= enddate]
        return md

    md_joined = None
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for md in executor.map(_retrieve_md, symbols):
            if md_joined is None:
                md_joined = md
            elif md is not None:
                md_joined = md_joined.join(md, how="outer")
            else:
                continue

        # futures = []
        # for symbol in symbols:
        #     futures.append(executor.submit(_retrieve_md, symbol, observable, startdate, enddate, cachedir))
        # for future in concurrent.futures.as_completed(futures):
        #     md = future.result()
        #     if md_joined is None:
        #         md_joined = md
        #     elif md is not None:
        #         md_joined = md_joined.join(md, how='outer')
        #     else:
        #         continue
    return md_joined

    md_joined = None
    # download one by one to avoid rejection
    for symbol in symbols:
        file_path = None if cachedir is None else cachedir / f"md_{symbol}.csv".lower()
        md = None
        if file_path is not None and file_path.exists():
            md = pd.read_csv(file_path, index_col="Date")
        else:
            md = yf.download(symbol)
            if file_path is not None:
                md.to_csv(file_path)

        if md.empty:
            continue

        md.index = pd.to_datetime(md.index)
        md = md[[observable]]
        md.columns = [symbol.upper()]
        if startdate is not None:
            md = md[md.index >= startdate]
        if enddate is not None:
            md = md[md.index <= enddate]

        if md_joined is None:
            md_joined = md
        else:
            md_joined = md_joined.join(md, how="outer")

    return md_joined



"""
The module provides users with high-level usage of the library and it's
implemented to :
- conduct standard anlaysis from asset recommendations, portfolio buildup
  to performance analysis;
- with options to specify:
    - pool of assets
    - "historical" period as well as "future period"
    - market data and hence portfolio update frequency
"""

from .dataaccessor import marketdata
from .portfolios.blacklitterman import BlackLittermanPortfolio
from .portfolios.longonly import LongOnlyPortfolio
from .portfolios.minvariance import MinVariancePortfolio
from .portfolios.riskbudgeting import RiskBudgetingPortfolio
from .utils import math
from .reporting import reports
from . import portfolios

import numpy as np
import pandas as pd
import datetime as dt


def asset_recommendations(today, lookbackdays, frequency, assetrank, symbols=None):
    """Asset recommendations:
    Extract out assets as recommendation from a large asset pool.
    This is mainly based on correlations.

    Args:
        today (str): assumed T0 date in string with format 'yyyy-mm-dd'
        lookbackdays (int): calendar days to date back from today
        frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y
        assetrank (int): number of tops from correlation pairs
        symbols (array-like, optional): a list of valid symbols. Defaults to None.

    Returns:
        pd.DataFrame: recommended assets with related attributes.
    """

    # Get all assets candidates when None
    asset_info = marketdata.get_asset_info()
    if symbols is None:
        symbols = list(asset_info.Symbol)

    hist_end = dt.datetime.strptime(today, "%Y-%m-%d")
    hist_start = hist_end - dt.timedelta(days=lookbackdays)

    # covariance of equilibrium excess returns
    rfr = marketdata.get_md_rfr(hist_start, hist_end, frequency=frequency)
    asset_close = marketdata.get_md_observable(
        symbols, "Adj Close", hist_start, hist_end, frequency=frequency
    )

    # Remove assets that contain too few historical data
    asset_close_count = asset_close.count()
    close_row_count = asset_close.shape[0]
    for index, value in asset_close_count.items():
        if value < close_row_count:
            asset_close.drop(columns=index, inplace=True)

    asset_returns = np.log(asset_close / asset_close.shift(1)).dropna()
    asset_returns_and_rfr = asset_returns.join(rfr).fillna(method="backfill")
    excess_returns = asset_returns_and_rfr.subtract(
        asset_returns_and_rfr["RFR"], axis="index"
    ).iloc[:, :-1]

    correlations = excess_returns.corr()
    smalls, zeros, bigs = math.correlation_rankings(correlations, assetrank)
    recommended_assets = set(np.array([i[0] for i in zeros]).flatten())
    print(f"{len(recommended_assets)} assets are recommended:")
    print(recommended_assets)

    excess_returns = excess_returns[recommended_assets]
    correlations = excess_returns.corr()

    recommended_assets = asset_info[asset_info.Symbol.isin(recommended_assets)]
    reports.create_asset_recommendation_report(
        correlations, assetrank, recommended_assets
    )
    return recommended_assets


def asset_selections(symbols, today, lookbackdays, frequency):
    """Asset selections:
    Conduct some deep analysis of selected assets, which should be normally from
    preliminary recommendation exercise.

    Args:
        symbols (array-like): a list of valid symbols
        today (str): assumed T0 date in string with format 'yyyy-mm-dd'
        lookbackdays (int): calendar days to date back from today
        frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y

    Returns:
        pd.DataFrame: covariance related stats among selected assets.
    """

    hist_end = dt.datetime.strptime(today, "%Y-%m-%d")
    hist_start = hist_end - dt.timedelta(days=lookbackdays)

    # covariance of equilibrium excess returns
    rfr = marketdata.get_md_rfr(hist_start, hist_end, frequency=frequency)
    asset_close = marketdata.get_md_observable(
        symbols, "Adj Close", hist_start, hist_end, frequency=frequency
    )

    # Remove assets that contain too few historical data
    asset_close_count = asset_close.count()
    close_row_count = asset_close.shape[0]
    for index, value in asset_close_count.items():
        if value < close_row_count:
            asset_close.drop(columns=index, inplace=True)

    asset_returns = np.log(asset_close / asset_close.shift(1)).dropna()
    asset_returns_and_rfr = asset_returns.join(rfr).fillna(method="backfill")
    excess_returns = asset_returns_and_rfr.subtract(
        asset_returns_and_rfr["RFR"], axis="index"
    ).iloc[:, :-1]

    naive_name = "Naive"
    naive_cov = math.get_cov(excess_returns, robust=False)
    naive_cov_eigs = sorted(np.linalg.eig(naive_cov)[0], reverse=True)
    naive_cov_det = np.linalg.det(naive_cov)
    naive_cov_inv = np.linalg.inv(naive_cov)
    naive_cov_inv_det = np.linalg.det(naive_cov_inv)
    naive_series = pd.Series(
        {
            "Eigen Values": naive_cov_eigs,
            "$\lambda_{max}/\lambda_{min}$": abs(
                naive_cov_eigs[-1] / naive_cov_eigs[0]
            ),
            "Determinant": naive_cov_det,
            "Inv Determinant": naive_cov_inv_det,
        },
        name=naive_name,
    )

    robust_name = "Robust"
    robust_cov = math.get_cov(excess_returns, robust=True)
    robust_cov_eigs = sorted(np.linalg.eig(robust_cov)[0], reverse=True)
    robust_cov_det = np.linalg.det(robust_cov)
    robust_cov_inv = np.linalg.inv(robust_cov)
    robust_cov_inv_det = np.linalg.det(robust_cov_inv)
    robust_series = pd.Series(
        {
            "Eigen Values": robust_cov_eigs,
            "$\lambda_{max}/\lambda_{min}$": abs(
                robust_cov_eigs[-1] / robust_cov_eigs[0]
            ),
            "Determinant": robust_cov_det,
            "Inv Determinant": robust_cov_inv_det,
        },
        name=robust_name,
    )

    cov_stats = pd.DataFrame([naive_series, robust_series]).T
    reports.create_asset_selection_report(cov_stats, excess_returns.corr())

    cov_stats = cov_stats.append(
        pd.Series({"Naive": naive_cov, "Robust": robust_cov}, name="Cov. Matrix")
    )
    return cov_stats


def naive_robust_cov_difference(
    symbols,
    equilweights,
    today,
    lookbackdays,
    lookaheaddays,
    views,
    frequency,
    riskaversion,
):
    """Naive vs robust covariance estimates:
    Check difference between using naive and robust convariance methods.

    Args:
        symbols (array-like): a list of valid symbols
        equilweights (array-like): weighting vector for equilibrium portfolio
        today (str): assumed T0 date in string with format 'yyyy-mm-dd'
        lookbackdays (int): calendar days to date back from today
        lookaheaddays (int): calendar days to date forward from today
        views (pd.DataFrame): including info on P and Q, e.g:
            pd.DataFrame({'view1': [1, 0, 0.01], 'view2': [0, 1, -0.05]}, index=['s1', 's2', 'Q'])
        frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y
        riskaversion (float): risk aversion to define investor's appetite

    Returns:
        None: N/A
    """

    comparison = _PortfolioComparison(
        symbols, equilweights, today, lookbackdays, lookaheaddays, frequency
    )

    P = views.values[:-1].T
    Q = views.values[-1].reshape(-1, 1)
    excess_returns = comparison.ins_excess_returns_
    equil_weights = comparison.equil_weights_

    def get_pfl_weights(robust):
        equil_cov = math.get_cov(excess_returns, robust)
        pfl = BlackLittermanPortfolio(
            equil_cov,
            equil_weights.values.reshape(-1, 1),
            P,
            Q,
            riskaversion=riskaversion,
        )
        if riskaversion is None:
            print(
                f"Implied risk aversion (robust?->{robust}): ", pfl.get_risk_aversion()
            )
        return pfl.get_weights(version="v1")

    pfl_name = "BL with Naive Covariances"
    pfl_weights = get_pfl_weights(False)
    pfl_weights = pd.Series(
        pfl_weights.reshape(-1),
        index=equil_weights.index,
        name="BL with Naive Covariances",
    )
    comparison.append_pfl(pfl_weights)

    pfl_name = "BL with Robust Covariances"
    pfl_weights = get_pfl_weights(True)
    pfl_weights = pd.Series(
        pfl_weights.reshape(-1),
        index=equil_weights.index,
        name="BL with Robust Covariances",
    )
    comparison.append_pfl(pfl_weights)

    comparison.create_report()


def blacklitterman_riskaversion(
    symbols,
    equilweights,
    today,
    lookbackdays,
    lookaheaddays,
    views,
    frequency,
    riskaversions,
):
    """Analyse impacts from risk aversions:

    Args:
        symbols (array-like): a list of valid symbols
        equilweights (array-like): weighting vector for equilibrium portfolio
        today (str): assumed T0 date in string with format 'yyyy-mm-dd'
        lookbackdays (int): calendar days to date back from today
        lookaheaddays (int): calendar days to date forward from today
        views (pd.DataFrame): including info on P and Q, e.g:
            pd.DataFrame({'view1': [1, 0, 0.01], 'view2': [0, 1, -0.05]}, index=['s1', 's2', 'Q'])
        frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y
        riskaversion (float): risk aversion to define investor's appetite

    Returns:
        None: N/A
    """

    comparison = _PortfolioComparison(
        symbols, equilweights, today, lookbackdays, lookaheaddays, frequency
    )

    P = views.values[:-1].T
    Q = views.values[-1].reshape(-1, 1)
    excess_returns = comparison.ins_excess_returns_
    equil_weights = comparison.equil_weights_

    equil_cov = math.get_cov(excess_returns)

    for ra in riskaversions:
        pfl = BlackLittermanPortfolio(
            equil_cov, equil_weights.values.reshape(-1, 1), P, Q, ra
        )
        pfl_name = rf"BL Portfolio $\lambda$={ra}"
        pfl_weights = pfl.get_weights(version="v1")
        pfl_weights = pd.Series(pfl_weights.reshape(-1), index=symbols, name=pfl_name)
        comparison.append_pfl(pfl_weights)

    comparison.create_report()


def blacklitterman_with_constraints(
    symbols,
    equilweights,
    today,
    lookbackdays,
    lookaheaddays,
    views,
    frequency,
    riskaversion,
):
    """Analyse impacts from different constraints:

    Args:
        symbols (array-like): a list of valid symbols
        equilweights (array-like): weighting vector for equilibrium portfolio
        today (str): assumed T0 date in string with format 'yyyy-mm-dd'
        lookbackdays (int): calendar days to date back from today
        lookaheaddays (int): calendar days to date forward from today
        views (pd.DataFrame): including info on P and Q, e.g:
            pd.DataFrame({'view1': [1, 0, 0.01], 'view2': [0, 1, -0.05]}, index=['s1', 's2', 'Q'])
        frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y
        riskaversion (float): risk aversion to define investor's appetite

    Returns:
        None: N/A
    """

    comparison = _PortfolioComparison(
        symbols, equilweights, today, lookbackdays, lookaheaddays, frequency
    )

    P = views.values[:-1].T
    Q = views.values[-1].reshape(-1, 1)
    excess_returns = comparison.ins_excess_returns_
    equil_weights = comparison.equil_weights_

    equil_cov = math.get_cov(excess_returns)
    pfl = BlackLittermanPortfolio(
        equil_cov, equil_weights.values.reshape(-1, 1), P, Q, riskaversion=riskaversion
    )

    pfl_name = "BL Portfolio: Unconstrained"
    pfl_weights = pfl.get_weights(version="v1")
    pfl_weights = pd.Series(pfl_weights.reshape(-1), index=symbols, name=pfl_name)
    comparison.append_pfl(pfl_weights)

    pfl_name = "BL Portfolio: Equal Weights"
    pfl_weights = [1.0 / len(symbols)] * len(symbols)
    pfl_weights = pd.Series(pfl_weights, index=symbols, name=pfl_name)
    comparison.append_pfl(pfl_weights)

    pfl_name = "BL Portfolio: Min Variance"
    pfl_weights = MinVariancePortfolio(pfl.get_covariances()).get_weights()
    pfl_weights = pd.Series(pfl_weights.reshape(-1), index=symbols, name=pfl_name)
    comparison.append_pfl(pfl_weights)

    budget = 1.5  # can borrow 50% of balance at maximum

    pfl_name = "BL Portfolio: Long Only"
    pfl_weights = LongOnlyPortfolio(
        pfl.get_covariances(), pfl.get_returns(), budget
    ).get_weights()
    pfl_weights = pd.Series(pfl_weights.reshape(-1), index=symbols, name=pfl_name)
    comparison.append_pfl(pfl_weights)

    pfl_name = "BL Portfolio: ERC"
    symbol_count = len(symbols)
    risk_budgets = np.array([1.0 / symbol_count] * symbol_count).reshape(-1, 1)
    pfl_weights = RiskBudgetingPortfolio(
        pfl.get_covariances(), pfl.get_returns(), budget, riskbudgets=risk_budgets
    ).get_weights()
    pfl_weights = pd.Series(pfl_weights.reshape(-1), index=symbols, name=pfl_name)
    comparison.append_pfl(pfl_weights)

    comparison.create_report()


class _PortfolioComparison(object):
    """Class for store and comapring a number of portfolios:

    Args:
        object ([type]): N/A
    """

    def __init__(
        self, symbols, equilweights, today, lookbackdays, lookaheaddays, frequency
    ):
        """Constructor to initialise equilibrium portfolio plus:
        - excess returns
        - market returns for internal usage

        Args:
            symbols (array-like): a list of valid symbols
            equilweights (array-like): weighting vector for equilibrium portfolio
            today (str): assumed T0 date in string with format 'yyyy-mm-dd'
            lookbackdays (int): calendar days to date back from today
            lookaheaddays (int): calendar days to date forward from today
            frequency (str): available ones - 1D, 1W, 1M, 3M and 1Y
        """

        today = dt.datetime.strptime(today, "%Y-%m-%d")
        hist_date = today - dt.timedelta(days=lookbackdays)
        future_date = today + dt.timedelta(days=lookaheaddays)
        rfr = marketdata.get_md_rfr(hist_date, future_date, frequency=frequency)
        asset_close = marketdata.get_md_observable(symbols, frequency=frequency)
        asset_close = asset_close[
            (asset_close.index >= hist_date) & (asset_close.index <= future_date)
        ]
        asset_returns = np.log(asset_close / asset_close.shift(1)).dropna()
        asset_returns_and_rfr = asset_returns.join(rfr).fillna(method="backfill")
        excess_returns = asset_returns_and_rfr.subtract(
            asset_returns_and_rfr["RFR"], axis="index"
        ).iloc[:, :-1]
        ins_excess_returns = excess_returns[excess_returns.index <= today]

        equil_name = "Equilibrium Portfolio"
        equil_weights = pd.Series(equilweights.ravel(), symbols, name=equil_name)
        equil_weights = equil_weights / equil_weights.sum()
        equil_returns = (
            asset_returns_and_rfr @ portfolios.balance_weights_with_cash(equil_weights)
        ).iloc[:, 0]
        equil_returns.name = equil_name

        self._today = today
        self._asset_returns_and_rfr = asset_returns_and_rfr
        self._weights = equil_weights.to_frame()
        self._returns = equil_returns.to_frame()
        self._equil_name = equil_name

        self.ins_excess_returns_ = ins_excess_returns
        self.equil_weights_ = equil_weights
        self.equil_returns_ = equil_returns

    def append_pfl(self, pflweights):
        """Append a portfolio represented by allocation

        Args:
            pflweights (array-like): weighting vector for appended portfolio

        Returns:
            _PortfolioComparison: the same instance
        """

        pfl_returns = (
            self._asset_returns_and_rfr
            @ portfolios.balance_weights_with_cash(pflweights)
        ).iloc[:, 0]
        pfl_returns.name = pflweights.name
        self._weights = self._weights.join(pflweights)
        self._returns = self._returns.join(pfl_returns)
        return self

    def create_report(self):
        """Create report"""

        return reports.create_overall_report(
            self._returns,
            today=self._today,
            benchmarkreturns=self._returns[self._equil_name],
            weights=self._weights,
            benchmarkweights=self._weights[self._equil_name],
        )



from .. import show

import unittest as ut
import pandas as pd


class ShowTests(ut.TestCase):
    def test_performance_stats(self):
        returns = pd.DataFrame(
            {
                "Portfolio 1": [0.05, 0.07, -0.01, 0.1, 0.01],
                "Portfolio 2": [0.033, -0.054, 0.3, 0.005, -0.032],
            },
            index=[
                "2011-07-01",
                "2012-07-01",
                "2013-07-01",
                "2014-07-01",
                "2015-07-01",
            ],
        )
        returns.index = pd.to_datetime(returns.index)
        actual = show.performance_stats(returns)
        expected = pd.DataFrame(
            {
                "Portfolio 1": [
                    "2011-07-01",
                    "2015-07-01",
                    5,
                    0,
                    -0.01,
                    0.1,
                    0.044,
                    0.0445,
                    -1.5810,
                    -0.0159,
                    1.2461,
                    0.9888,
                    0.1563,
                    -0.006,
                ],
                "Portfolio 2": [
                    "2011-07-01",
                    "2015-07-01",
                    5,
                    0,
                    -0.054,
                    0.3,
                    0.0504,
                    0.1435,
                    3.9449,
                    1.9399,
                    1.3284,
                    0.3513,
                    0.2629,
                    -0.0496,
                ],
            },
            [
                "Start Date",
                "End Date",
                "Count of Returns",
                "Count of NaN Returns",
                "Min of Returns",
                "Max of Returns",
                "Mean of Returns",
                "Volatility of Returns",
                "Kurtosis of Returns",
                "Skew of Returns",
                "Cumulative Returns",
                "Sharpe Ratio",
                "Max Drawdown",
                "Value at Risk (95%)",
            ],
        )
        pd.testing.assert_frame_equal(expected, actual)


if __name__ == "__main__":
    ut.main()



import seaborn as sns
import pandas as pd
from tabulate import tabulate
import numpy as np
from scipy.stats import norm
from . import math

import matplotlib.pyplot as plt

plt.style.use("fivethirtyeight")


def performance_stats(returns, riskfreerate=0):
    """Plot performance statistics:
    Check some standard characteristics of included portfolios, each of which
    represented by a return time series.

    Args:
        returns (pd.DataFrame): rows represent time series; columns are for scoped portfolios
        riskfreerate (float, optional): a static risk free rate. Defaults to 0.

    Returns:
        pd.DataFrame: Basic stats included, where rows are chracteristics and columns are portfolios
    """
    mean = returns.mean()
    volatility = returns.std()
    sharpe_ratio = (mean - riskfreerate) / volatility
    cumsum = returns.cumsum().apply(np.exp)
    minmum = returns.min()
    maximum = returns.max()
    cumsum_min = cumsum.min()
    cumsum_max = cumsum.max()
    max_drawdown = (cumsum_max - cumsum_min) / cumsum_max
    stats = {
        "Start Date": returns.index.min().strftime("%Y-%m-%d"),
        "End Date": returns.index.max().strftime("%Y-%m-%d"),
        "Count of Returns": returns.count(),
        "Count of NaN Returns": returns.isna().sum(),
        "Min of Returns": minmum.round(4),
        "Max of Returns": maximum.round(4),
        "Mean of Returns": mean.round(4),
        "Volatility of Returns": volatility.round(4),
        "Kurtosis of Returns": returns.kurtosis().round(4),
        "Skew of Returns": returns.skew().round(4),
        "Cumulative Returns": cumsum.max().round(4),
        "Sharpe Ratio": sharpe_ratio.round(4),
        "Max Drawdown": max_drawdown.round(4),
        "Value at Risk (95%)": returns.quantile(0.05).round(4),
    }
    stats = pd.DataFrame(stats).T
    print("Table: Portfolio Performance Statistics")
    print_as_table(stats)
    return stats


def asset_allocations(weights, benchmarkweights):
    """Plot asset allocation:
    Compare asset allocations among portfolios.

    Args:
        weights (pd.DataFrame): rows are weights and columns are portfolios
        benchmarkweights (pd.Series): weightings for benchmark portfolio
    """
    weights_with_bk = (
        weights
        if benchmarkweights.name in weights.columns
        else weights.join(benchmarkweights)
    )
    plt.figure()
    (weights_with_bk * 100).T.plot(kind="barh", stacked=True, figsize=(14, 6))
    plt.xlabel("Weights (%)"), plt.ylabel("Portfolios")
    plt.title("Comparison of Asset Allocation among Portfolios")
    plt.show()
    return


def returns_aggs(returns):
    """Plot some aggregated stats derived from returns, including:
    - return KDE
    - return candle box

    Args:
        returns (pd.DataFrame): rows represent time series; columns are for scoped portfolios
    """
    returns_ptg = returns * 100
    plt.figure()
    returns_ptg.plot(kind="kde", figsize=(14, 6), linewidth=0.8, alpha=0.7)
    plt.xlabel("Returns (%)"), plt.ylabel("Probability")
    plt.title("Return Distribution (KDE)")
    plt.show()

    plt.figure()
    returns_ptg.plot(kind="box", figsize=(14, 6))
    plt.xlabel("Portfolios"), plt.ylabel("Returns (%)")
    plt.title("Return Distribution (Box)")
    plt.show()


def rolling_characteristics(returns, riskfreerate=0):
    """Plot rollable characteristics derived from returns, including:
    - raw return ts
    - volatility ts
    - SR ts

    Args:
        returns (pd.DataFrame): rows represent time series; columns are for scoped portfolios
        riskfreerate (float, optional): a static risk free rate. Defaults to 0.
    """
    returns_ptg = returns * 100
    plt.figure()
    returns_ptg.plot(figsize=(14, 6), linewidth=0.8, alpha=0.7)
    plt.xlabel("Date"), plt.ylabel("Returns (%)")
    plt.title("Return Time Series")
    plt.show()

    plt.figure()
    returns_ptg.rolling(10).std().plot(figsize=(14, 6), linewidth=0.8, alpha=0.7)
    plt.xlabel("Date"), plt.ylabel("Volatility (%)")
    plt.title("Rolling Volatility (window=10)")

    plt.figure()
    sharpe_ratio = (
        returns_ptg.rolling(10).mean() - riskfreerate
    ) / returns_ptg.rolling(10).std()
    sharpe_ratio.plot(figsize=(14, 6), linewidth=0.8, alpha=0.7)
    plt.xlabel("Date"), plt.ylabel("Sharpe Ratio (%)")
    plt.title("Rolling Sharpe Ratio (window=10)")
    plt.show()


def profit_and_loss(returns, today, benchmarkreturns=None):
    """Plot P&L in time:
    - P&L
    - Drawdowns

    Args:
        returns (pd.DataFrame): rows represent time series; columns are for scoped portfolios
        today ('str'): assumed T0 date in string with format 'yyyy-mm-dd'
        benchmarkreturns (pd.Series): return ts for benchmark portfolio

    Raises:
        ValueError: expect at least 2 portfolios are included in returns at the moment.
    """
    returns_with_bk = (
        returns
        if benchmarkreturns.name in returns.columns
        else returns.join(benchmarkreturns)
    )
    returns_with_bk = returns_with_bk[returns_with_bk.index > today]
    nrows = returns_with_bk.shape[0]
    for i in range(nrows):
        if returns_with_bk.index[i] <= pd.to_datetime(today):
            returns_with_bk.iloc[i, :-1] = returns_with_bk.iloc[i, -1]

    pnl = returns_with_bk.cumsum().apply(np.exp) * 100.0
    pnl.iloc[0, :] = 100.0
    pnl.plot(linewidth=1, figsize=(14, 6))

    pfl_count = pnl.shape[1]
    row_count = int(np.ceil(pfl_count / 2.0))
    fig, axes = plt.subplots(row_count, 2, sharex=True, sharey=True)
    fig.tight_layout(h_pad=2)
    axes.reshape(row_count, 2)
    drawdown_table_data = pd.DataFrame()
    overall_worst_count = -np.inf
    if pfl_count == 1:
        raise ValueError(
            f"At 2 portfolios are expected the moment - (Only {pnl.columns[0]} is available.)"
        )
    if row_count == 1:
        for column_idx in range(2):
            s = pnl.iloc[:, column_idx]
            ax = axes[column_idx]
            ax.set_title(s.name)
            s.plot(ax=ax, linewidth=0.8, figsize=(14, 6))
            smin = s.min()
            smax = s.max()
            drawdowns = math.get_drawdowns(s)
            drawdown_keys = sorted(drawdowns, key=drawdowns.get, reverse=True)
            worst_count = min(len(drawdown_keys), 5)
            overall_worst_count = max(worst_count, overall_worst_count)
            drawdown_keys = drawdown_keys[:worst_count]
            for idx, key in zip(range(len(drawdown_keys)), drawdown_keys):
                xs = s.index[key[0]]
                xe = s.index[key[2]]
                ax.fill_between([xs, xe], smin, smax, alpha=0.3)
                drawdown_table_data[(s.name, idx)] = [
                    f"{drawdowns[key]:.2%}",
                    xs.strftime("%Y-%m-%d"),
                    s.index[key[1]].strftime("%Y-%m-%d"),
                    xe.strftime("%Y-%m-%d"),
                    (xe - xs).days,
                ]
    else:
        for row_idx in range(row_count):
            for column_idx in range(2):
                pnl_column_idx = row_idx * 2 + column_idx
                ax = axes[row_idx, column_idx]
                if pnl_column_idx >= pfl_count:
                    fig.delaxes(ax)
                    continue
                s = pnl.iloc[:, pnl_column_idx]
                ax.set_title(s.name)
                s.plot(ax=ax, linewidth=0.8, figsize=(14, 6))
                smin = s.min()
                smax = s.max()
                drawdowns = math.get_drawdowns(s)
                drawdown_keys = sorted(drawdowns, key=drawdowns.get, reverse=True)
                worst_count = min(len(drawdown_keys), 5)
                overall_worst_count = max(worst_count, overall_worst_count)
                drawdown_keys = drawdown_keys[:worst_count]
                for idx, key in zip(range(len(drawdown_keys)), drawdown_keys):
                    xs = s.index[key[0]]
                    xe = s.index[key[2]]
                    ax.fill_between([xs, xe], smin, smax, alpha=0.3)
                    drawdown_table_data[(s.name, idx)] = [
                        f"{drawdowns[key]:.2%}",
                        xs.strftime("%Y-%m-%d"),
                        s.index[key[1]].strftime("%Y-%m-%d"),
                        xe.strftime("%Y-%m-%d"),
                        (xe - xs).days,
                    ]
    fig.suptitle(f"{overall_worst_count} Worst Drawdowns")
    plt.show()
    drawdown_table_data = drawdown_table_data.T
    drawdown_table_data.columns = [
        "Net drawdown in %",
        "Peak Date",
        "Valley date",
        "Recovery date",
        "Duration (Days)",
    ]
    print(f"Table: {worst_count} Worst Drawdowns")
    print_as_table(drawdown_table_data)


def correlations(matrix, rankcount):
    """Plot correlation matrix as heatmap:

    Args:
        matrix (pd.DataFrame): both rows and columns are assets
        rankcount (int): number of tops from correlation pairs
    """
    smalls, zeros, bigs = math.correlation_rankings(matrix, rankcount)
    rank_data = pd.DataFrame(
        {
            f"Biggest {rankcount}": bigs,
            f"Smallest {rankcount}": smalls,
            f"{rankcount} closest to 0": zeros,
        }
    )
    print_as_table(rank_data)

    plt.figure(figsize=[14, 14])
    sns.heatmap(matrix.round(2), annot=True, cmap="autumn", fmt="g")
    plt.title("Correlation Matrix")
    plt.ylabel("Assets"), plt.xlabel("Assets")
    plt.show()


def print_as_table(data):
    """Plot tabulars from data:
    3 tables are plots under different format:
    - simple psql-like display (readable)
    - html format that can be interpreted by markdown
    - latex format for latex writings

    Args:
        data (pd.DataFrame): Any pd.DataFrame object.
    """
    print(tabulate(data, headers="keys", tablefmt="psql"))
    print(tabulate(data, headers="keys", tablefmt="html"))
    print(tabulate(data, headers="keys", tablefmt="latex"))


def verbose_print(verbose=True, *args, **kwargs):
    """Wrap original python print:
    This allows us to decide whether to actual action on print

    Args:
        verbose (bool, optional): should print if true. Defaults to True.
    """
    if not verbose:
        return
    print(*args, **kwargs)



"""noofolio root init file implemented to some key APIs for users to access,
which includes:
- Key functions/classes, e.g.: noofolio.BlackLittermanPortfolio
- sub-modules, e.g.: noofolio.marketdata
"""

from .portfolio import Portfolio
from . import screen


from .portfolios.blacklitterman import BlackLittermanPortfolio
from .portfolios.longonly import LongOnlyPortfolio
from .portfolios.minvariance import MinVariancePortfolio
from .portfolios.riskbudgeting import RiskBudgetingPortfolio

from .reporting import reports
from .utils import math, show


__version__ = "0.1.0"



from .. import marketdata

import unittest as ut
from pathlib import Path


class MarketDataTests(ut.TestCase):
    def test_get_md_observable_single(self):
        close = marketdata.get_md_observable(
            "JD", cachedir=Path(__file__).resolve().parent / "data"
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "2014-05-22")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 1666)

    def test_get_md_observable_multi(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"], cachedir=Path(__file__).resolve().parent / "data"
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-12")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 10099)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_md_observable_eod(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"],
            cachedir=Path(__file__).resolve().parent / "data",
            frequency="1D",
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-12")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 10099)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_md_observable_eow(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"],
            cachedir=Path(__file__).resolve().parent / "data",
            frequency="1W",
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-12")
        self.assertEqual(close.index[1].strftime("%Y-%m-%d"), "1980-12-19")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 2091)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_md_observable_eom(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"],
            cachedir=Path(__file__).resolve().parent / "data",
            frequency="1M",
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-31")
        self.assertEqual(close.index[1].strftime("%Y-%m-%d"), "1981-01-30")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 481)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_md_observable_eoq(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"],
            cachedir=Path(__file__).resolve().parent / "data",
            frequency="3M",
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-31")
        self.assertEqual(close.index[1].strftime("%Y-%m-%d"), "1981-03-31")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 161)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_md_observable_eoy(self):
        close = marketdata.get_md_observable(
            ["JD", "AAPL"],
            cachedir=Path(__file__).resolve().parent / "data",
            frequency="1Y",
        )
        self.assertEqual(close.index.min().strftime("%Y-%m-%d"), "1980-12-31")
        self.assertEqual(close.index.max().strftime("%Y-%m-%d"), "2020-12-31")
        self.assertEqual(len(close), 41)
        self.assertListEqual(list(close.columns), ["JD", "AAPL"])

    def test_get_symbols_from_cache(self):
        actual = marketdata.get_symbols_from_cache(
            cachedir=Path(__file__).resolve().parent / "data"
        )
        self.assertListEqual(actual, ["aapl", "jd"])


if __name__ == "__main__":
    ut.main()



import numpy as np


def balance_weights_with_cash(riskyassetweights):
    """get weights that consider the portion borrowed on risk free rate:

    Args:
        riskyassetweights (np.array): weights for all risky assets

    Returns:
        np.array: vector that appends risk-free part
    """
    weights_sum = np.sum(riskyassetweights)
    cash_weight = 1.0 - weights_sum
    weights = np.append(riskyassetweights, cash_weight).reshape(-1, 1)
    return weights



from ..utils import show

import numpy as np
import pandas as pd


def create_overall_report(
    returns, today, benchmarkreturns=None, weights=None, benchmarkweights=None
):
    """create a generic report:

    Args:
        returns (pd.DataFrame): rows represent time series; columns are for scoped portfolios
        today ('str'): assumed T0 date in string with format 'yyyy-mm-dd'
        benchmarkreturns (pd.Series): return ts for benchmark portfolio
        weights (pd.DataFrame, optional): rows are weights and columns are portfolios. Defaults to None.
        benchmarkweights (pd.Series, optional): weightings for benchmark portfolio. Defaults to None.
    """
    print("=" * 128)
    print("-" * 44, "Portfolio Construction: Overall Report", "-" * 44)
    print("_" * 128)

    look_ahead_rets = returns[returns.index > today]
    show.performance_stats(look_ahead_rets)
    if weights is not None:
        show.asset_allocations(weights, benchmarkweights)
    show.returns_aggs(look_ahead_rets)
    show.rolling_characteristics(look_ahead_rets)
    show.profit_and_loss(returns, today, benchmarkreturns)

    print("_" * 128)


def create_asset_recommendation_report(correlations, assetrank, recommendedassets):
    """create asset recommendation report:

    Args:
        correlations (pd.DataFrame): correlation matrix
        assetrank (int): Number of tops that we would like to extract out
        recommendedassets (pd.DataFrame): assets with basic information on each
    """
    print("=" * 128)
    print("-" * 37, "Portfolio Construction: Asset Recommendation Report", "-" * 38)
    print("_" * 128)

    show.correlations(correlations, assetrank)
    recommendedassets.sort_values(by="Market Cap", ascending=False).plot(
        kind="bar", x="Symbol", y="Market Cap", figsize=(14, 6), logy=True
    )
    show.print_as_table(
        recommendedassets[["Symbol", "Name", "Country", "Market Cap", "Industry"]]
    )

    print("_" * 128)


def create_asset_selection_report(covstats, correlations):
    """Create report on selected assets:

    Args:
        covstats (pd.DataFrame): some basic statistics around covariances
        correlations (pd.DataFrame): covariance matrix
    """
    print("=" * 128)
    print("-" * 40, "Portfolio Construction: Selected Asset Report", "-" * 41)
    print("_" * 128)

    show.correlations(correlations, np.inf)
    # TODO: eigen values
    show.print_as_table(covstats)

    naive_eigs = covstats[covstats.index == "Eigen Values"]["Naive"][0]
    robust_eigs = covstats[covstats.index == "Eigen Values"]["Robust"][0]
    pd.DataFrame({"Naive": naive_eigs, "Robust": robust_eigs}).plot(figsize=(14, 6))

    print("_" * 128)



from .baseportfolio import BasePortfolio

import numpy as np
import scipy as sp


class LongOnlyPortfolio(BasePortfolio):
    """Class for modelling long-only portfolios:

    Args:
        BasePortfolio (BasePortfolio): inheriting for some commonly used functions, e.g.: get_weights
    """

    def __init__(self, covariances, returns, budget=1.0):
        """Constructor for passing in covariance matrix and returns of assets in portfolio:

        Args:
            covariances (np.array): covariance matrix
            returns (np.array): column vector for individual asset returns
            budget (float, optional): capital usable for investment, which is expressed relative term.
                1.0 represents full balance in one's account. Defaults to 1.0.
        """
        # inheriting from base
        super().__init__()
        self._covariances = covariances
        self._returns = returns

        self._budget = budget
        self._risk_aversion = None

    def get_weights(self, riskaversion=1.0):
        """Override function from base:
        - if weights have been calculated once and risk aversion hasn't changed, use cached ones;
        - otherwise, recalculate the weights

        Args:
            riskaversion (float, optional): risk aversion to reflect one's appetite. Defaults to 1.0.

        Returns:
            np.array: column vector calculated from the model
        """
        if self._weights is not None and self._risk_aversion == riskaversion:
            return self._weights

        def objective_func(weights):
            w = weights.reshape(-1, 1)

            risk = (w.T @ self._covariances @ w * 0.5 * riskaversion)[0][0]
            ret = (w.T @ self._returns)[0][0]
            return risk - ret

            sig_weights = self._covariances @ w
            to_return = sig_weights * riskaversion - self._returns
            return np.sum(to_return**2.0)

        n = len(self._returns)
        res = sp.optimize.minimize(
            objective_func,
            [1.0 / float(n)] * n,  # x0
            method="SLSQP",  # sequential least square quadratic programming
            bounds=sp.optimize.Bounds(0, np.inf),  # 0 <= x
            constraints=sp.optimize.LinearConstraint(
                np.ones((1, n)), -np.inf, self._budget
            ),  # 1^T x <= budget
            options={"disp": True, "ftol": 1e-11},
        )
        weights = res.x.reshape(n, 1)
        self._weights = weights
        return weights



import numpy as np
from sklearn.covariance import LedoitWolf


def get_cov(data, robust=True, verbose=True):
    """Compute covariance matrix from samples:
    - naive: simply by the numpy.cov implementation
    - robust: Ledoit-Wolf method implemented in sklearn

    Args:
        data (matrix-like): sample data where a row a 1 sample, and columns represent variables
        robust (bool, optional): use robust method. Defaults to True.
        verbose (bool, optional): print available details. Defaults to True.

    Returns:
        [type]: [description]
    """
    if robust:
        estimator = LedoitWolf().fit(data)
        if verbose:
            print("LedoitWolf:")
            print("    shrinkage: ", estimator.shrinkage_)
            print("    precision: ", estimator.precision_)
            print("    location: ", estimator.location_)
        return estimator.covariance_

    return np.cov(data, rowvar=False)


def get_drawdowns(data):
    """Compute draw downs over a time series:

    Args:
        data (array-like): the time series

    Returns:
        dict: (min_index, max_index) -> drawdown mappings
    """
    emin = data[0]
    emax = data[0]
    imin = 0
    imax = 0
    n = len(data)
    drawdowns = {}
    downed = False
    for i, e in zip(range(n), data):
        if e < emin:
            emin = e
            imin = i
            downed = True
        elif e >= emax and downed:
            drawdowns[(imax, imin, i)] = (emax - emin) / emax
            emax = e
            emin = e
            imax = i
            downed = False
        elif e > emax:
            emax = e
            emin = e
            imax = i
        else:
            continue

    if (emin < emax) and downed:
        drawdowns[(imax, imin, n - 1)] = (emax - emin) / emax
    return drawdowns


def correlation_rankings(matrix, rankcount):
    """Extract tops from correlation matrix:
    - most correlated pairs
    - most uncorrelated pairs
    - most inversely correlated pairs.

    Args:
        matrix (pd.DataFrame): the correlation matrix
        rankcount ([type]): first N from the ranked tops

    Returns:
        tuple: 3 entries, each is a list of (pairs, correlation_value)
    """
    corr_dict = {}
    asset_count = len(matrix)
    rankcount = min(asset_count, rankcount)
    for row_idx in range(asset_count):
        for column_idx in range(row_idx):
            corr_dict[
                (matrix.index[row_idx], matrix.columns[column_idx])
            ] = matrix.iloc[row_idx, column_idx]
    corr_dict_keys = sorted(corr_dict, key=corr_dict.get)
    smalls = [(key, corr_dict[key]) for key in corr_dict_keys[:rankcount]]
    bigs = [(key, corr_dict[key]) for key in corr_dict_keys[-1 : -rankcount - 1 : -1]]
    corr_dict_abs = sorted(corr_dict.items(), key=lambda item: np.abs(item[1]))
    zeros = corr_dict_abs[:rankcount]
    return smalls, zeros, bigs



"""Stock Screener

"""
import logging
import numpy as np
import pandas as pd
from scipy import stats
from sklearn.preprocessing import Normalizer, MinMaxScaler
from sklearn.neighbors import LocalOutlierFactor


_LOGGER = logging.getLogger("noofolio")


class Screener(object):
    def __init__(self, data=None, path=None) -> None:
        self._rates = None
        self._data = data
        pass

    def run(self):
        if self._rates:
            return self._rates

        data = self._data
        feature_summary = pd.concat(
            [data.dtypes, data.nunique() / data.notna().sum()],
            axis=1,
            keys=["dtypes", "nunique"],
        )
        feature_summary["nnan"] = data.isna().sum() / data.shape[0]
        feature_summary["numeric?"] = np.where(
            (feature_summary["nunique"] > 0.5)
            & (feature_summary["dtypes"] == "float64"),
            True,
            False,
        )
        # feature_summary.loc["stats_peg", "numeric?"] = True
        feature_summary.loc["stats_roe", "numeric?"] = True
        feature_summary.loc["stats_roa", "numeric?"] = True
        data_cleaned = data.copy()
        # remove delisted firms
        data_cleaned = data_cleaned[data_cleaned["delisted"] == False]

        # remove stats_roe == -99999.99
        data_cleaned[data_cleaned["stats_roe"] == -99999.99] = np.nan
        # remove stats_roe == 0.0
        data_cleaned[data_cleaned["stats_roe"] == 0.0] = np.nan
        # remove stats_roa == 0.0
        data_cleaned[data_cleaned["stats_roa"] == 0.0] = np.nan
        # remove pnl_grossmargin == inf or -inf
        data_cleaned[abs(data_cleaned["pnl_grossmargin"]) == np.inf] = np.nan

        # # only take fundamentals within 360 days asofdate
        # data_cleaned = data_cleaned[
        #     pd.to_datetime(data_cleaned["fundamental_date"]) + dt.timedelta(days=360)
        #     > as_of_date
        # ]

        nonan_data = data_cleaned[
            list(feature_summary[feature_summary["numeric?"] == True].index)
        ].copy()
        nonan_data["ticker"] = data_cleaned["ticker"] + "." + data_cleaned["exchange"]
        nonan_data.dropna(inplace=True)

        # normalizing
        normlizer = Normalizer()
        cf_columns = ["cf_operating", "cf_investing", "cf_financing"]
        cf_transformed = normlizer.fit_transform(nonan_data[cf_columns])
        x_train = nonan_data[
            feature_summary[feature_summary["numeric?"] == True].index
        ].copy()
        x_train[cf_columns] = cf_transformed

        # scaling
        columns_to_scale = [
            c for c in x_train.columns if "bs_" in c or "pnl_" in c or "stats_" in c
        ]
        desc = x_train.describe()
        for c in columns_to_scale:
            mini = desc.loc["min", c]
            maxi = desc.loc["max", c]
            scaler = MinMaxScaler((0.01, maxi - mini + 0.01))
            transformed = scaler.fit_transform(x_train[c].values.reshape(-1, 1))
            x_train.loc[:, c] = transformed
        # x_train["bs_leverage"] = stats.boxcox(x_train["bs_leverage"])[0]
        # x_train["stats_ep"] = stats.boxcox(x_train["stats_ep"])[0]
        # x_train["stats_roe"] = stats.boxcox(x_train["stats_roe"])[0]
        # x_train["stats_bp"] = stats.boxcox(x_train["stats_bp"])[0]
        scaler = MinMaxScaler((0, 1))
        x_train[x_train.columns] = scaler.fit_transform(x_train)

        # outlier detection
        outlier_detector = LocalOutlierFactor()
        y_train_predict = outlier_detector.fit_predict(x_train)

        return list(nonan_data.ticker.iloc[np.where(y_train_predict == -1)])



from ..riskbudgeting import RiskBudgetingPortfolio

import unittest
import numpy as np


class RiskBudgetingPortfolioTestsWithElectiveRBPort1(unittest.TestCase):

    def setUp(self):
        correlations = np.array(
            [[1.00, 0.50, 0.25], [0.50, 1.00, 0.60], [0.25, 0.60, 1.00]]
        )
        std_dev = np.diagflat([0.02, 0.03, 0.01])
        covariances = std_dev @ correlations @ std_dev
        returns = np.array([[0.005], [0.003], [0.002]])
        weights = np.array([[0.5203], [0.1439], [0.3358]])
        self._portfolio = RiskBudgetingPortfolio(
            covariances, returns, 1.0, weights=weights
        )

    def test_get_valueatrisk(self):
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.90), -0.0152, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.95), -0.0206, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.99), -0.0306, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.995), -0.0343, 4)

    def test_get_expectedshortfall(self):
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.90), -0.0222, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.95), -0.0267, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.99), -0.0356, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.995), -0.0390, 4)


@unittest.skip("to correct according to baseline later")
class RiskBudgetingPortfolioTestsWithElectiveRBPort2(unittest.TestCase):

    def setUp(self):
        correlations = np.array(
            [[1.00, 0.50, 0.25], [0.50, 1.00, 0.60], [0.25, 0.60, 1.00]]
        )
        std_dev = np.diagflat([0.02, 0.03, 0.01])
        covariances = std_dev @ correlations @ std_dev
        returns = np.array([[0.005], [0.003], [0.002]])
        # weights = np.array([[0.7305], [-0.2021], [0.4716]])
        weights = np.array([[488.0], [-135.0], [315]])
        self._portfolio = RiskBudgetingPortfolio(
            covariances, returns, 1.0, weights=weights
        )

    def test_get_valueatrisk(self):
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.90), -0.0568, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.95), -0.0745, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.99), -0.1076, 4)
        self.assertAlmostEqual(self._portfolio.get_valueatrisk(0.995), -0.1198, 4)

    def test_get_expectedshortfall(self):
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.90), -0.0798, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.95), -0.0948, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.99), -0.1241, 4)
        self.assertAlmostEqual(self._portfolio.get_expectedshortfall(0.995), -0.1352, 4)


class RiskBudgetingPortfolioTestsWithElectiveRBExample7(unittest.TestCase):

    def setUp(self):
        correlations = np.array(
            [[1.00, 0.80, 0.50], [0.80, 1.00, 0.30], [0.50, 0.30, 1.00]]
        )
        std_dev = np.diagflat([0.30, 0.20, 0.15])
        covariances = std_dev @ correlations @ std_dev
        returns = np.array([[0.0], [0.0], [0.0]])
        weights = np.array([[0.5], [0.2], [0.3]])
        self._portfolio = RiskBudgetingPortfolio(
            covariances, returns, 1.0, weights=weights
        )

    def test_get_variance(self):
        self.assertAlmostEqual(self._portfolio.get_variance(), 0.043555, 6)

    def test_get_riskcontribution(self):
        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(returnpercentage=False),
            np.array([[0.1470], [0.0333], [0.0285]]),
            2e-3,
        )
        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(returnpercentage=True),
            np.array([[0.7043], [0.1593], [0.1364]]),
            3e-4,
        )

        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(
                by="var", confidence=0.99, returnpercentage=False
            ),
            np.array([[-0.3419], [-0.0774], [-0.0662]]),
            6e-4,
        )
        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(
                by="var", confidence=0.99, returnpercentage=True
            ),
            np.array([[0.7043], [0.1593], [0.1364]]),
            3e-4,
        )

        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(
                by="es", confidence=0.99, returnpercentage=False
            ),
            np.array([[-0.3917], [-0.0886], [-0.0759]]),
            6e-4,
        )
        np.testing.assert_allclose(
            self._portfolio.get_riskcontribution(
                by="es", confidence=0.99, returnpercentage=True
            ),
            np.array([[0.7043], [0.1593], [0.1364]]),
            3e-4,
        )


class RiskBudgetingPortfolioTestsWithElectiveRBExample11(unittest.TestCase):

    def setUp(self):
        correlations = np.array(
            [[1.00, 0.80, 0.50], [0.80, 1.00, 0.30], [0.50, 0.30, 1.00]]
        )
        std_dev = np.diagflat([0.30, 0.20, 0.15])
        covariances = std_dev @ correlations @ std_dev
        returns = np.array([[0.10], [0.05], [0.09]])
        risk_budgets = np.array([[0.5], [0.2], [0.3]])
        self._portfolio = RiskBudgetingPortfolio(
            covariances, returns, 1.0, riskbudgets=risk_budgets
        )

    def test_get_weights(self):
        np.testing.assert_allclose(
            self._portfolio.get_weights(confidence=0.95),
            np.array([[0.2821], [0.1968], [0.5211]]),
            atol=5e-4,
        )
        np.testing.assert_allclose(
            self._portfolio.get_weights(confidence=0.99),
            np.array([[0.29], [0.2026], [0.5074]]),
            atol=5e-4,
        )


if __name__ == "__main__":
    unittest.main()



from .baseportfolio import BasePortfolio

import numpy as np
import scipy as sp


class RiskBudgetingPortfolio(BasePortfolio):
    """Class for modelling risk-budgeting portfolios:

    Args:
        BasePortfolio (BasePortfolio): inheriting for some commonly used functions, e.g.: get_weights
    """

    def __init__(
        self,
        covariances,
        returns,
        budget,
        riskbudgets=None,
        weights=None,
        by="es",
        confidence=0.99,
    ):
        """Constructor for initialising member variables

        Args:
            covariances (np.array): convariance matrix
            returns (np.array): column vector for returns of assets comprising of the portfolio
            budget (float): maximum number of captital for usage. 1 standard the full balance in one's account;
                1.5 means 50% of one's own money is borrowed
            riskbudgets (np.array, optional): column vector for risk assigned to each asset.
                It's in relative term. Defaults to None.
            weights (np.array, optional): column vector of the portfolio which is exclusive with
                riskbudgets above. Defaults to None.
            by (str, optional): risk measure option. Defaults to 'es'.
            confidence (float, optional): confidence level applied to risk measure. Defaults to 0.99.
        """
        # inheriting from base
        super().__init__()
        self._covariances = covariances
        self._returns = returns
        self._weights = weights

        # specific to class itself
        self._budgets = budget
        self._risk_budgets = riskbudgets
        self._confidence = confidence
        self._risk_measure = by

    def get_weights(self, by="es", confidence=0.99):
        """Get weights for the portfolio:
        - if risk budget allocation is not define, return whatever is from its base get_weights
        - if weights have been calculated and no change in risk measure yet, return cached _weights
        - otherwise, recalculate weights from risk budgeting definition
        """
        if self._risk_budgets is None:
            return super().get_weights()

        if (
            self._risk_measure == by
            and self._confidence == confidence
            and self._weights is not None
        ):
            return self._weights

        def objective_func(init_weights):
            sig_weights = self._covariances @ init_weights
            marginal_risk = sig_weights / np.sqrt(init_weights @ sig_weights)
            if by == "var":
                c = sp.stats.norm.ppf(confidence)
                marginal_risk = self._returns.T - c * marginal_risk
            elif by == "es":
                c = sp.stats.norm.pdf(sp.stats.norm.ppf(confidence)) / (
                    1.0 - confidence
                )
                marginal_risk = self._returns.T - c * marginal_risk
            elif by != "volatility":
                raise ValueError(f"risk measure [{by}] is not recognized.")
            rc = init_weights * marginal_risk
            rc = rc / np.sum(rc)
            return np.sum((rc - self._risk_budgets.T) ** 2.0)

        n = len(self._returns)
        res = sp.optimize.minimize(
            objective_func,
            [1.0 / float(n)] * n,  # x0
            method="SLSQP",  # sequential least square quadratic programming
            bounds=sp.optimize.Bounds(0, np.inf),  # 0 <= x long only
            constraints=sp.optimize.LinearConstraint(
                [1] * n, -np.inf, self._budgets
            ),  # 1^T x <=budget
        )
        weights = np.atleast_2d(res.x).reshape(n, 1)
        self._weights = weights
        self._risk_measure = by
        self._confidence = confidence
        return weights



from .baseportfolio import BasePortfolio

import numpy as np


class MinVariancePortfolio(BasePortfolio):
    """Class for modelling min-variance portfolios:

    Args:
        BasePortfolio (BasePortfolio): inheriting for some commonly used functions.
    """

    def __init__(self, covariances):
        """Constructor where convariances are passed in

        Args:
            covariances (np.array): covariance matrix
        """
        # inheriting from base
        super().__init__()
        self._covariances = covariances

    def get_weights(self):
        """override function from base"""
        n = len(self._covariances)
        _covariancesinv = np.linalg.inv(self._covariances)
        ones = np.ones((n, 1))
        a = _covariancesinv @ ones
        return a / (ones.T @ a)



import numpy as np
import unittest

from ..minvariance import MinVariancePortfolio


class MinVariancePortfolioTestsWithM2L2(unittest.TestCase):
    """
    docstring here
        :param unittest2.TestCase
    """

    _minvar = None

    def setUp(self):
        covariances = np.array(
            [
                [0.0049, 0.00672, 0.0105, 0.0168],
                [0.00672, 0.0144, 0.0252, 0.036],
                [0.0105, 0.0252, 0.09, 0.144],
                [0.0168, 0.036, 0.144, 0.36],
            ]
        )
        self._minvar = MinVariancePortfolio(covariances)

    def test_get_weights(self):
        actual = self._minvar.get_weights()
        np.testing.assert_allclose(
            [[1.274886723], [-0.263112728], [0.016339421], [-0.028113415]], actual
        )

    def test_covariances_(self):
        actual = self._minvar._covariances
        np.testing.assert_allclose(
            np.array(
                [
                    [0.0049, 0.00672, 0.0105, 0.0168],
                    [0.00672, 0.0144, 0.0252, 0.036],
                    [0.0105, 0.0252, 0.09, 0.144],
                    [0.0168, 0.036, 0.144, 0.36],
                ]
            ),
            actual,
        )


if __name__ == "__main__":
    unittest.main()



import abc
import numpy as np
from scipy.stats import norm


class BasePortfolio(abc.ABC):
    """Abstract class that can be inherited by any concrete portfolio model in theory:

    Args:
        ABC (ABC): force abstract class nature
    """
    def __init__(self):
        """Constructor that initialise some protected member variables
        """
        # portfolio attributes
        self._covariances = None
        self._returns = None
        self._weights = None

        # calculated
        self._variance = None

    # @abc.abstractclassmethod # disable such restriction for now
    def get_covariances(self, **kwargs):
        """Base method to get covariance matrix for the portfolio:

        Returns:
            **kwargs: actual parameters might differ in concrete portfolio classes
        """
        return self._covariances

    def get_weights(self, **kwargs):
        """Base method to get weights for the portfolio:

        Returns:
            **kwargs: actual parameters might differ in concrete portfolio classes
        """
        return self._weights

    def get_returns(self, **kwargs):
        """Base method to get returns for the portfolio:

        Returns:
            **kwargs: actual parameters might differ in concrete portfolio classes
        """
        return self._returns

    def get_variance(self):
        """Get variance for the portfolio
        """
        if self._variance is None:
            if self._covariances is None:
                raise ValueError(
                    'Portfolio attribute [covariance] is not set yet.')
            weights = self.get_weights()
            self._variance = (weights.T @ self._covariances @ weights)[0][0]
        return self._variance

    def get_stddev(self):
        """Get standard deviation for the portfolio
        """
        return np.sqrt(self.get_variance())

    def get_valueatrisk(self, confidence):
        """Get VaR for the portfolio
        """
        weights = self.get_weights()
        c = norm.ppf(confidence)
        return (self._returns.T @ weights - c * self.get_stddev())[0][0]

    def get_expectedshortfall(self, confidence):
        """Get ES for the portfolio
        """
        weights = self.get_weights()
        c = norm.pdf(norm.ppf(confidence)) / (1.0 - confidence)
        return (self._returns.T @ weights - c * self.get_stddev())[0][0]

    def get_riskcontribution(self, by='volatility',
                             confidence=None, returnpercentage=True):
        """Get risk contributions from each asset in the portfolio
        """
        weights = self.get_weights()
        sig_weights = self._covariances @ weights
        marginal_risk = sig_weights / np.sqrt(weights.T @ sig_weights)[0][0]

        if by == 'var':
            c = norm.ppf(confidence)
            marginal_risk = self._returns - c * marginal_risk
        elif by == 'es':
            c = norm.pdf(norm.ppf(confidence)) / (1.0 - confidence)
            marginal_risk = self._returns - c * marginal_risk
        elif by != 'volatility':
            raise ValueError(f'risk measure [{by}] is not recognized.')
        rc = weights * marginal_risk

        return rc / np.sum(rc) if returnpercentage else rc



from .. import math

import unittest as ut
import numpy as np
import pandas as pd


class MathTests(ut.TestCase):
    def test_get_cov(self):
        # TODO
        return

    def test_get_drawdowns(self):
        data = [1] * 10
        self.assertDictEqual(math.get_drawdowns(data), {})

        data = [1, 0.2, 0.3, 0.05, 1]
        self.assertDictEqual(math.get_drawdowns(data), {(0, 3, 4): 0.95})

        data = [1, 0.2, 0.3, 2.0, 0.05, 1]
        self.assertDictEqual(
            math.get_drawdowns(data), {(0, 1, 3): 0.8, (3, 4, 5): 0.975}
        )

        data = [1, 0.2, 0.3, 2.0, 2.3, 2.5, 0.5, 1.5]
        self.assertDictEqual(math.get_drawdowns(data), {(0, 1, 3): 0.8, (5, 6, 7): 0.8})

        data = [1, 0.2, 0.3, 2.0, 0.05, 3.5]
        self.assertDictEqual(
            math.get_drawdowns(data), {(0, 1, 3): 0.8, (3, 4, 5): 0.975}
        )
        self.assertDictEqual(
            math.get_drawdowns(pd.Series(data)), {(0, 1, 3): 0.8, (3, 4, 5): 0.975}
        )

    def test_correlation_rankings(self):
        samples = pd.DataFrame(
            {
                "column1": [0.1, 1.5, 1.3, 1.0, 0.9, 0.8],
                "column2": [0.2, 0.5, 0.3, 0.0, 0.5, 0.3],
                "column3": [-0.1, -2.5, 3.3, 0.1, 0.9, 0.8],
            }
        )
        s, z, b = math.correlation_rankings(samples.corr(), 2)

        columns = sorted(set(np.array([i[0] for i in s]).flatten()))
        self.assertSequenceEqual(columns, ["column1", "column2", "column3"])

        self.assertSequenceEqual(s[0][0], ("column3", "column2"))
        self.assertSequenceEqual(s[1][0], ("column3", "column1"))
        self.assertAlmostEqual(s[0][1], -0.19119537521843286)
        self.assertAlmostEqual(s[1][1], -0.0557977069592031)

        self.assertSequenceEqual(z[0][0], ("column3", "column1"))
        self.assertSequenceEqual(z[1][0], ("column3", "column2"))
        self.assertAlmostEqual(z[0][1], -0.0557977069592031)
        self.assertAlmostEqual(z[1][1], -0.19119537521843286)

        self.assertSequenceEqual(b[0][0], ("column2", "column1"))
        self.assertSequenceEqual(b[1][0], ("column3", "column1"))
        self.assertAlmostEqual(b[0][1], 0.36991501890585793)
        self.assertAlmostEqual(b[1][1], -0.0557977069592031)


if __name__ == "__main__":
    ut.main()



import logging
import numpy as np
from copy import deepcopy

from .portfolios.minvariance import MinVariancePortfolio
from .utils import math

logger = logging.getLogger(__name__)


def _folio_backtest(closes, portfolio):
    tickers = closes.columns
    returns = closes.copy()
    returns = returns / returns.iloc[0]
    returns["Portfolio"] = returns @ portfolio.weights
    return returns


def portfolio_optimise(portfolio, model, asofdate, periodstart, periodend, rfr=0):
    """TODO:
    - handle rfr
    - functional-like so that no arg is expected to have change in value
    - for portfolio, introduce inplace option

    Args:
        portfolio (_type_): _description_
        model (_type_): _description_
        asofdate (_type_): _description_
        periodstart (_type_): _description_
        periodend (_type_): _description_
        rfr (int, optional): _description_. Defaults to 0.

    Raises:
        Exception: _description_

    Returns:
        _type_: _description_
    """
    pfl_optimised = deepcopy(portfolio)
    if pfl_optimised.timeseries is None:
        timeseries = None
        for ticker in pfl_optimised.tickers:
            # TODO: retire yfanance
            ts = None
            # ts = access.via.yahoo.get_daily_ts_for_ticker(
            #     ticker, periodstart, periodend
            # )[["trading_date", "adj_close"]]
            if timeseries is None:
                timeseries = ts
            else:
                timeseries = timeseries.merge(ts, on="trading_date", how="inner")
        timeseries.set_index("trading_date", inplace=True)
        timeseries.columns = pfl_optimised.tickers
        pfl_optimised.timeseries = timeseries

    if pfl_optimised.weights is None:
        ticker_count = len(pfl_optimised.tickers)
        pfl_optimised.weights = [1.0 / ticker_count] * ticker_count

    # portfolio re-weighting
    hist_daily_rets = pfl_optimised.timeseries[
        pfl_optimised.timeseries.index < asofdate
    ].copy()
    hist_daily_rets = np.log(hist_daily_rets / hist_daily_rets.shift(1)).dropna()
    pfl_model = None
    if model == "MinVariance":
        pfl_model = MinVariancePortfolio(math.get_cov(hist_daily_rets))
    else:
        raise Exception(f"Defined portfolio model [{model}] not recognised")
    pfl_optimised.weights = pfl_model.get_weights()

    return pfl_optimised
